#include "main.h"

static struct TS3Functions ts3Functions;

#ifdef WINDOWS
#define _strcpy(dest, destSize, src) strcpy_s(dest, destSize, src)
#define snprintf sprintf_s
#else
#define _strcpy(dest, destSize, src) { strncpy(dest, src, destSize-1); dest[destSize-1] = '\0'; }
#endif

#define PATH_BUFSIZE 512
#define COMMAND_BUFSIZE 128

static char* pluginCommandID = NULL;

/*********************************** Required functions ************************************/
/*
 * If any of these required functions is not implemented, TS3 will refuse to load the plugin
 */

/* Unique name identifying this plugin */
const char* ts3plugin_name()
{
    return "TS3MassMover";
}

/* Plugin version */
const char* ts3plugin_version()
{
    return "0.2";
}

/* Plugin API version. Must be the same as the clients API major version, else the plugin fails to load. */
int ts3plugin_apiVersion()
{
	return 8;
}

/* Plugin author */
const char* ts3plugin_author()
{
    return "Stefan1200";
}

/* Plugin description */
const char* ts3plugin_description()
{
    return "This plugin adds short chat commands to move all players from one channel! Visit http://www.stefan1200.de";
}

/* Set TeamSpeak 3 callback functions */
void ts3plugin_setFunctionPointers(const struct TS3Functions funcs)
{
    ts3Functions = funcs;
}

/*
 * Custom code called right after loading the plugin. Returns 0 on success, 1 on failure.
 * If the function returns 1 on failure, the plugin will be unloaded again.
 */
int ts3plugin_init()
{
    return 0;  /* 0 = success, 1 = failure */
}

/* Custom code called right before the plugin is unloaded */
void ts3plugin_shutdown()
{
    /* Your plugin cleanup code here */
    printf("TS3MassMover PLUGIN: shutdown\n");

	/* Free pluginCommandID if we registered it */
	if(pluginCommandID)
	{
		free(pluginCommandID);
		pluginCommandID = NULL;
	}
}

/*
* If the plugin wants to use plugin commands, it needs to register a command ID. This function will be automatically called after
* the plugin was initialized. This function is optional. If you don't use plugin commands, the function can be omitted.
* Note the passed commandID parameter is no longer valid after calling this function, so you *must* copy it and store it in the plugin.
*/
void ts3plugin_registerPluginCommandID(const char* commandID)
{
	const size_t sz = strlen(commandID) + 1;
	pluginCommandID = (char*)malloc(sz);
	memset(pluginCommandID, 0, sz);
	_strcpy(pluginCommandID, sz, commandID);  /* The commandID buffer will invalidate after existing this function */
	printf("PLUGIN: registerPluginCommandID: %s\n", pluginCommandID);
}

/* Plugin command keyword. Return NULL or "" if not used. */
const char* ts3plugin_commandKeyword()
{
	return "m";
}

/* Plugin processes console command. Return 0 if plugin handled the command, 1 if not handled. */
int ts3plugin_processCommand(uint64 serverConnectionHandlerID, const char* command)
{
	char buf[COMMAND_BUFSIZE];
	char *s, *param1 = NULL, *param2 = NULL, *password = const_cast<char *>("");
	int i = 0;
	anyID myID;
	enum { CMD_NONE = 0, CMD_MHELP, CMD_MH, CMD_MAH, CMD_MT, CMD_MAT } cmd = CMD_NONE;
#ifdef WINDOWS
	char* context = NULL;
#endif

	printf("PLUGIN: process command: '%s'\n", command);

	_strcpy(buf, COMMAND_BUFSIZE, command);
#ifdef WINDOWS
	s = strtok_s(buf, " ", &context);
#else
	s = strtok(buf, " ");
#endif
	while(s != NULL)
	{
		if(i == 0)
		{
		    if(!strcmp(s, "help"))
			{
				cmd = CMD_MHELP;
			}
			else if(!strcmp(s, "h"))
			{
				cmd = CMD_MH;
			}
			else if(!strcmp(s, "ah"))
			{
				cmd = CMD_MAH;
			}
			else if(!strcmp(s, "t"))
			{
				cmd = CMD_MT;
			}
			else if(!strcmp(s, "at"))
			{
				cmd = CMD_MAT;
			}
		}
		else if(i == 1)
		{
			param1 = s;
		}
		else
		{
			param2 = s;
		}
#ifdef WINDOWS
		s = strtok_s(NULL, " ", &context);
#else
		s = strtok(NULL, " ");
#endif
		i++;
	}

	switch(cmd)
	{
		case CMD_NONE:
			return 1;  /* Command not handled by plugin */
		case CMD_MT:
			if(param1)
			{
				uint64 channelID = (uint64)atoi(param1);
				password = const_cast<char *>(param2 ? param2 : "");

				/* Get own clientID */
				if(ts3Functions.getClientID(serverConnectionHandlerID, &myID) != ERROR_ok)
				{
					ts3Functions.logMessage("Error querying client ID", LogLevel_ERROR, "TS3MassMover Plugin", serverConnectionHandlerID);
					break;
				}

				uint64 currentChannelID;
				if(ts3Functions.getChannelOfClient(serverConnectionHandlerID, myID, &currentChannelID) != ERROR_ok)
				{
					ts3Functions.logMessage("Error querying channel ID", LogLevel_ERROR, "TS3MassMover Plugin", serverConnectionHandlerID);
					break;
				}

				anyID* channelClientList;
				if(ts3Functions.getChannelClientList(serverConnectionHandlerID, currentChannelID, &channelClientList) != ERROR_ok)
				{
					ts3Functions.logMessage("Error querying channel client list", LogLevel_ERROR, "TS3MassMover Plugin", serverConnectionHandlerID);
					break;
				}

                int clientType;
				for(i=0; channelClientList[i]; i++)
				{
                    if(ts3Functions.getClientVariableAsInt(serverConnectionHandlerID, channelClientList[i], CLIENT_TYPE, &clientType) != ERROR_ok)
                    {
                        continue;
                    }
                    if (clientType == 1)
                    {
                        continue;
                    }

                    /* Move clients to specified channel */
                    if(ts3Functions.requestClientMove(serverConnectionHandlerID, channelClientList[i], channelID, password, NULL) != ERROR_ok)
                    {
                        ts3Functions.logMessage("Error requesting client move", LogLevel_ERROR, "TS3MassMover Plugin", serverConnectionHandlerID);
                    }
				}

				ts3Functions.freeMemory(channelClientList);
			}
			else
			{
				ts3Functions.printMessageToCurrentTab("Missing channel ID parameter.");
			}
			break;
		case CMD_MAT:
			if(param1)
			{
				uint64 channelID = (uint64)atoi(param1);
				password = const_cast<char *>(param2 ? param2 : "");

				anyID* clientList;
				if(ts3Functions.getClientList(serverConnectionHandlerID, &clientList) != ERROR_ok)
				{
					ts3Functions.logMessage("Error querying client list", LogLevel_ERROR, "TS3MassMover Plugin", serverConnectionHandlerID);
					break;
				}

				anyID* channelClientList;
				if(ts3Functions.getChannelClientList(serverConnectionHandlerID, channelID, &channelClientList) != ERROR_ok)
				{
					ts3Functions.logMessage("Error querying channel client list", LogLevel_ERROR, "TS3MassMover Plugin", serverConnectionHandlerID);
                    ts3Functions.freeMemory(clientList);
					break;
				}

				//char buffer [60];

                int clientType;
				for(i=0; clientList[i]; i++)
				{
                    if(ts3Functions.getClientVariableAsInt(serverConnectionHandlerID, clientList[i], CLIENT_TYPE, &clientType) != ERROR_ok)
                    {
                        continue;
                    }
                    if (clientType == 1)
                    {
                        continue;
                    }

                    if (isClientInList(channelClientList, &clientList[i]))
                    {
                        continue;
                    }

                    // Debug
                    //sprintf (buffer, "Debug: Moving client ID %d to channel ID %d", clientList[i], channelID);
                    //ts3Functions.printMessageToCurrentTab(buffer);

                    /* Move clients to specified channel */
                    if(ts3Functions.requestClientMove(serverConnectionHandlerID, clientList[i], channelID, password, NULL) != ERROR_ok)
                    {
                        ts3Functions.logMessage("Error requesting client move", LogLevel_ERROR, "TS3MassMover Plugin", serverConnectionHandlerID);
                    }
				}

                ts3Functions.freeMemory(clientList);
				ts3Functions.freeMemory(channelClientList);
			}
			else
			{
				ts3Functions.printMessageToCurrentTab("Missing channel ID parameter.");
			}
			break;
		case CMD_MH:
			if(param1)
			{
				uint64 channelID = (uint64)atoi(param1);
				password = const_cast<char *>(param2 ? param2 : "");

				/* Get own clientID */
				if(ts3Functions.getClientID(serverConnectionHandlerID, &myID) != ERROR_ok)
				{
					ts3Functions.logMessage("Error querying client ID", LogLevel_ERROR, "TS3MassMover Plugin", serverConnectionHandlerID);
					break;
				}

				uint64 currentChannelID;
				if(ts3Functions.getChannelOfClient(serverConnectionHandlerID, myID, &currentChannelID) != ERROR_ok)
				{
					ts3Functions.logMessage("Error querying channel ID", LogLevel_ERROR, "TS3MassMover Plugin", serverConnectionHandlerID);
					break;
				}

				anyID* channelClientList;
				if(ts3Functions.getChannelClientList(serverConnectionHandlerID, channelID, &channelClientList) != ERROR_ok)
				{
					ts3Functions.logMessage("Error querying channel client list", LogLevel_ERROR, "TS3MassMover Plugin", serverConnectionHandlerID);
					break;
				}

                int clientType;
				for(i=0; channelClientList[i]; i++)
				{
                    if(ts3Functions.getClientVariableAsInt(serverConnectionHandlerID, channelClientList[i], CLIENT_TYPE, &clientType) != ERROR_ok)
                    {
                        continue;
                    }
                    if (clientType == 1)
                    {
                        continue;
                    }

                    /* Move clients to specified channel */
                    if(ts3Functions.requestClientMove(serverConnectionHandlerID, channelClientList[i], currentChannelID, password, NULL) != ERROR_ok)
                    {
                        ts3Functions.logMessage("Error requesting client move", LogLevel_ERROR, "TS3MassMover Plugin", serverConnectionHandlerID);
                    }
				}

				ts3Functions.freeMemory(channelClientList);
			}
			else
			{
				ts3Functions.printMessageToCurrentTab("Missing channel ID parameter.");
			}
			break;
        case CMD_MAH:
            password = const_cast<char *>(param1 ? param1 : "");

            /* Get own clientID */
            if(ts3Functions.getClientID(serverConnectionHandlerID, &myID) != ERROR_ok)
            {
                ts3Functions.logMessage("Error querying client ID", LogLevel_ERROR, "TS3MassMover Plugin", serverConnectionHandlerID);
                break;
            }

            uint64 currentChannelID;
            if(ts3Functions.getChannelOfClient(serverConnectionHandlerID, myID, &currentChannelID) != ERROR_ok)
            {
                ts3Functions.logMessage("Error querying channel ID", LogLevel_ERROR, "TS3MassMover Plugin", serverConnectionHandlerID);
                break;
            }

            anyID* clientList;
            if(ts3Functions.getClientList(serverConnectionHandlerID, &clientList) != ERROR_ok)
            {
                ts3Functions.logMessage("Error querying client list", LogLevel_ERROR, "TS3MassMover Plugin", serverConnectionHandlerID);
                break;
            }

            anyID* channelClientList;
            if(ts3Functions.getChannelClientList(serverConnectionHandlerID, currentChannelID, &channelClientList) != ERROR_ok)
            {
                ts3Functions.logMessage("Error querying channel client list", LogLevel_ERROR, "TS3MassMover Plugin", serverConnectionHandlerID);
                ts3Functions.freeMemory(clientList);
                break;
            }

            //char buffer [60];

            int clientType;
            for(i=0; clientList[i]; i++)
            {
                if(ts3Functions.getClientVariableAsInt(serverConnectionHandlerID, clientList[i], CLIENT_TYPE, &clientType) != ERROR_ok)
                {
                    continue;
                }
                if (clientType == 1)
                {
                    continue;
                }

                if (isClientInList(channelClientList, &clientList[i]))
                {
                    continue;
                }

                // Debug
                //sprintf (buffer, "Debug: Moving client ID %d to channel ID %d", clientList[i], currentChannelID);
                //ts3Functions.printMessageToCurrentTab(buffer);

                /* Move clients to specified channel */
                if(ts3Functions.requestClientMove(serverConnectionHandlerID, clientList[i], currentChannelID, password, NULL) != ERROR_ok)
                {
                    ts3Functions.logMessage("Error requesting client move", LogLevel_ERROR, "TS3MassMover Plugin", serverConnectionHandlerID);
                }
            }

            ts3Functions.freeMemory(clientList);
            ts3Functions.freeMemory(channelClientList);
            break;
		case CMD_MHELP:
			ts3Functions.printMessageToCurrentTab("TS3MassMover Plugin\n[b]/m h <channel id> [password of current channel][/b]\nMoves all players from given channel to your current channel!\n[b]/m ah [password of current channel][/b]\nMoves all players from server to your current channel!\n[b]/m t <channel id> [password of target channel][/b]\nMoves all players from your current channel to given channel!\n[b]/m at <channel id> [password of target channel][/b]\nMoves all players from server to given channel!");
	}

	return 0;  /* Plugin handled command */
}

bool isClientInList(anyID* clientList, anyID* clientID)
{
    for(int i=0; clientList[i]; i++)
    {
        if (clientList[i] == *clientID)
        {
            return true;
        }
    }

    return false;
}

/*
 * Plugin requests to be always automatically loaded by the TeamSpeak 3 client unless
 * the user manually disabled it in the plugin dialog.
 * This function is optional. If missing, no autoload is assumed.
 */
int ts3plugin_requestAutoload()
{
	return 1;  /* 1 = request autoloaded, 0 = do not request autoload */
}
